@@ -3,7 +3,12 @@ from ctypes.wintypes import DWORD
 
 from prompt_toolkit.key_binding.input_processor import KeyPress
 from prompt_toolkit.keys import Keys
-from prompt_toolkit.win32_types import EventTypes, KEY_EVENT_RECORD, INPUT_RECORD, STD_INPUT_HANDLE
+from prompt_toolkit.mouse_events import MouseEventTypes
+from prompt_toolkit.win32_types import EventTypes, KEY_EVENT_RECORD, MOUSE_EVENT_RECORD, INPUT_RECORD, STD_INPUT_HANDLE
+
+import msvcrt
+import os
+import sys
 
 __all__ = (
     'ConsoleInputReader',
@@ -68,6 +73,7 @@ class ConsoleInputReader(object):
         39: Keys.Right,
         40: Keys.Down,
 
+        45: Keys.Insert,
         46: Keys.Delete,
 
         # F-keys.
@@ -92,7 +98,20 @@ class ConsoleInputReader(object):
     RIGHT_CTRL_PRESSED = 0x0004
 
     def __init__(self):
-        self.handle = windll.kernel32.GetStdHandle(STD_INPUT_HANDLE)
+        self._fdcon = None
+
+        # When stdin is a tty, use that handle, otherwise, create a handle from
+        # CONIN$.
+        if sys.stdin.isatty():
+            self.handle = windll.kernel32.GetStdHandle(STD_INPUT_HANDLE)
+        else:
+            self._fdcon = os.open('CONIN$', os.O_RDWR | os.O_BINARY)
+            self.handle = msvcrt.get_osfhandle(self._fdcon)
+
+    def close(self):
+        " Close fdcon. "
+        if self._fdcon is not None:
+            os.close(self._fdcon)
 
     def read(self):
         """
@@ -130,6 +149,9 @@ class ConsoleInputReader(object):
                     if key_presses:
                         result.extend(key_presses)
 
+                elif type(ev) == MOUSE_EVENT_RECORD:
+                    result.extend(self._handle_mouse(ev))
+
         return result
 
     def _event_to_key_presses(self, ev):
@@ -186,9 +208,14 @@ class ConsoleInputReader(object):
 
         # Return result. If alt was pressed, prefix the result with an
         # 'Escape' key, just like unix VT100 terminals do.
+
+        # NOTE: Only replace the left alt with escape. The right alt key often
+        #       acts as altgr and is used in many non US keyboard layouts for
+        #       typing some special characters, like a backslash. We don't want
+        #       all backslashes to be prefixed with escape. (Esc-\ has a
+        #       meaning in E-macs, for instance.)
         if result:
-            meta_pressed = ev.ControlKeyState & self.LEFT_ALT_PRESSED or \
-                ev.ControlKeyState & self.RIGHT_ALT_PRESSED
+            meta_pressed = ev.ControlKeyState & self.LEFT_ALT_PRESSED
 
             if meta_pressed:
                 return [KeyPress(Keys.Escape, ''), result]
@@ -198,6 +225,27 @@ class ConsoleInputReader(object):
         else:
             return []
 
+    def _handle_mouse(self, ev):
+        """
+        Handle mouse events. Return a list of KeyPress instances.
+        """
+        FROM_LEFT_1ST_BUTTON_PRESSED = 0x1
+
+        result = []
+
+        # Check event type.
+        if ev.ButtonState == FROM_LEFT_1ST_BUTTON_PRESSED:
+            # On a key press, generate both the mouse down and up event.
+            for event_type in [MouseEventTypes.MOUSE_DOWN, MouseEventTypes.MOUSE_UP]:
+                data = ';'.join([
+                   event_type,
+                   str(ev.MousePosition.X),
+                   str(ev.MousePosition.Y)
+                ])
+                result.append(KeyPress(Keys.WindowsMouseEvent, data))
+
+        return result
+
 
 class raw_mode(object):
     """
@@ -210,7 +258,7 @@ class raw_mode(object):
     `raw_input` method of `.vt100_input`.
     """
     def __init__(self, fileno=None):
-        self.handle = windll.kernel32.GetStdHandle(-10)  # STD_INPUT_HANDLE
+        self.handle = windll.kernel32.GetStdHandle(STD_INPUT_HANDLE)
 
     def __enter__(self):
         # Remember original mode.
